<?php

namespace Myriad\Queue\Drivers\Google;

use Illuminate\Contracts\Encryption\Encrypter;
use Illuminate\Support\Str;
use RuntimeException;

trait InteractsWithPubSub
{

    /**
     * Encrypt and base64 encode the payload
     *
     * @param  mixed  $payload
     * @return string
     */
    public function constructRawPayload($payload): string
    {
        return base64_encode($this->getEncrypter()->encrypt($payload, false));
    }

    /**
     * Base64 decode and decrypt the payload
     *
     * @param  mixed  $payload
     * @return string
     */
    public function deconstructRawPayload($payload): string
    {
        return $this->getEncrypter()->decrypt(base64_decode($payload), false);
    }

    /**
     * Create a payload string from the given job and data.
     *
     * @param  mixed  $job
     * @param  mixed  $data
     * @param  array  $merge
     * @return string
     */
    public function constructPayload($job, $data = '', $merge = []): string
    {
        $payload = json_encode($this->constructPayloadArray($job, $data, $merge));
        if (JSON_ERROR_NONE !== json_last_error()) {
            throw new RuntimeException('Invalid JSON: '.json_last_error_msg());
        }
        return $payload;
    }

    /**
     * Deconstruct a payload
     *
     * @param  mixed  $payload
     * @return array
     */
    public function deconstructPayload($payload): array
    {
        return json_decode($this->deconstructRawPayload($payload), true);
    }

    /**
     * Create a payload array from the given job and data.
     *
     * @param  mixed  $job
     * @param  mixed  $data
     * @param  array  $merge
     * @return array
     */
    protected function constructPayloadArray($job, $data = '', $merge = []): array
    {
        if (is_array($job)) {
            $array = $job;
        } else {
            if (is_object($job)) {
                $array = $this->constructObjectPayload($job);
            } else {
                $array = $this->constructStringPayload($job, $data);
            }
        }
        return array_merge($array, $merge);
    }

    /**
     * Create a payload for an object-based queue handler.
     *
     * @param  mixed  $job
     * @return array
     */
    protected function constructObjectPayload($job): array
    {
        $uuid = (string) Str::uuid();

        return [
          'uuid' => $uuid,
          'id' => $uuid,
          'displayName' => method_exists($job, 'displayName') ? $job->displayName() : get_class($job),
          'job' => 'Illuminate\Queue\CallQueuedHandler@call',
          'maxTries' => method_exists($job, 'maxTries') ? $job->maxTries() : null,
          'timeout' => method_exists($job, 'timeout') ? $job->timeout() : null,
          'attempts' => method_exists($job, 'attempts') ? $job->attempts() : 1,
          'availableAt' => method_exists($job, 'availableAt') ? $job->availableAt() : null,
          'data' => [
            'commandName' => get_class($job),
            'command' => serialize(clone $job),
          ],
        ];
    }

    /**
     * Create a typical, string based queue payload array.
     *
     * @param  string  $job
     * @param  mixed  $data
     * @return array
     */
    protected function constructStringPayload($job, $data): array
    {
        $uuid = (string) Str::uuid();
        
        return [
          'uuid' => $uuid,
          'id' => $uuid,
          'displayName' => is_string($job) ? explode('@', $job)[0] : null, 'job' => $job,
          'maxTries' => null, 'timeout' => null, 'attempts' => 1, 'availableAt' => null, 'data' => $data,
        ];
    }

    /**
     * Get the encrypter
     *
     * @return Encrypter
     */
    public function getEncrypter(): Encrypter
    {
        $encrypter = $this->encrypter;
        if (empty($encrypter)) {
            $this->encrypter = app(Encrypter::class);
        }
        return $this->encrypter;
    }
}
