<?php

namespace Myriad\Queue\Drivers\Google;

use Carbon\Carbon;
use Exception;
use Google\Cloud\PubSub\Message;
use Google\Cloud\PubSub\Subscription;
use Illuminate\Container\Container;
use Illuminate\Contracts\Container\BindingResolutionException;
use Illuminate\Contracts\Queue\Job as JobContract;
use Illuminate\Queue\Jobs\Job;
use RuntimeException;

class PubSubQueueJob extends Job implements JobContract
{

    /**
     * @var PubSubQueue The queue
     */
    private $pubSubQueue;

    /**
     * @var Subscription The pub/sub subscription
     */
    private $subscription;

    /**
     * @var Message The message with encrypted payload
     */
    private $message;

    /**
     * @var array The message payload
     */
    private $payload;

    /**
     * PubSubQueueJob constructor.
     *
     * @param  PubSubQueue  $pubSubQueue
     * @param  Container  $container
     * @param  Subscription  $subscription
     * @param  string  $queue
     * @param  Message  $message
     * @param  array  $payload
     */
    public function __construct(
        PubSubQueue $pubSubQueue,
        Container $container,
        Subscription $subscription,
        string $queue,
        Message $message,
        array $payload
    ) {
        $this->pubSubQueue = $pubSubQueue;
        $this->container = $container;
        $this->subscription = $subscription;
        $this->queue = $queue;
        $this->message = $message;
        $this->payload = $payload;
        $this->connectionName = 'google';
    }

    /**
     * The IoC container instance.
     *
     * @var Container
     */
    protected $container;

    /**
     * Indicates if the job has been deleted.
     *
     * @var bool
     */
    protected $deleted = false;

    /**
     * Indicates if the job has been released.
     *
     * @var bool
     */
    protected $released = false;

    /**
     * Indicates if the job has failed.
     *
     * @var bool
     */
    protected $failed = false;

    /**
     * The name of the queue the job belongs to.
     *
     * @var string
     */
    protected $queue;

    /**
     * Fire the job.
     */
    public function fire(): void
    {
        $payload = $this->payload();

        list($class, $method) = CompatJobName::parse($payload['job']);

        $this->instance = $this->resolve($class);
        $this->instance->{$method}($this, $payload['data']);
    }

    /**
     * Renew the modify ack deadline for another n seconds.
     *
     * @param int $seconds
     */
    public function renew($seconds): void
    {
        if ($seconds > PubSubQueue::MAX_ACK_SECONDS) {
            throw new RuntimeException('Max renew time for Google Pub/Sub is '.PubSubQueue::MAX_ACK_SECONDS.' seconds.');
        }
        $this->subscription->modifyAckDeadline($this->message, $seconds);
    }

    /**
     * Delete the job from the queue.
     */
    public function delete(): void
    {
        $this->deleted = true;
        $this->subscription->acknowledge($this->message);
    }

    /**
     * Determine if the job has been deleted.
     *
     * @return bool
     */
    public function isDeleted(): bool
    {
        return $this->deleted;
    }

    /**
     * Release the job back into the queue.
     *
     * @param  int  $delay
     * @return void
     */
    public function release($delay = 0): void
    {
        $this->released = true;
        $this->pubSubQueue->pushRaw(
            $this->pubSubQueue->constructPayload(
                $this->payload(),
                null,
                ['availableAt' => Carbon::now()->addSeconds($delay)->timestamp, 'attempts' => $this->attempts() + 1]
            ),
            $this->queue
        );
        $this->subscription->acknowledge($this->message);
    }

    /**
     * Determine if the job was released back into the queue.
     *
     * @return bool
     */
    public function isReleased(): bool
    {
        return $this->released;
    }

    /**
     * Determine if the job has been deleted or released.
     *
     * @return bool
     */
    public function isDeletedOrReleased(): bool
    {
        return $this->isDeleted() || $this->isReleased();
    }

    /**
     * Determine if the job has been marked as a failure.
     *
     * @return bool
     */
    public function hasFailed(): bool
    {
        return $this->failed;
    }

    /**
     * Mark the job as "failed".
     */
    public function markAsFailed(): void
    {
        $this->failed = true;
    }

    /**
     * Process an exception that caused the job to fail.
     *
     * @param  \Throwable|null  $e
     * @return void
     */
    protected function failed($e)
    {
        $payload = $this->payload();

        [$class, $method] = CompatJobName::parse($payload['job']);

        if (method_exists($this->instance = $this->resolve($class), 'failed')) {
            $this->instance->failed($payload['data'], $e, $payload['uuid']);
        }
    }

    /**
     * Resolve the given class.
     *
     * @param  string  $class
     * @return mixed
     * @throws BindingResolutionException
     */
    protected function resolve($class)
    {
        return $this->container->make($class);
    }

    /**
     * Get the decoded body of the job.
     *
     * @return array
     */
    public function payload(): array
    {
        return $this->payload;
    }

    /**
     * The number of times to attempt a job.
     *
     * @return int|null
     */
    public function maxTries(): ?int
    {
        return data_get($this->payload(), 'maxTries');
    }

    /**
     * The number of seconds the job can run.
     *
     * @return int|null
     */
    public function timeout(): ?int
    {
        return data_get($this->payload(), 'timeout');
    }

    /**
     * Get the name of the queued job class.
     *
     * @return string
     */
    public function getName(): string
    {
        return $this->payload()['job'];
    }

    /**
     * Get the resolved name of the queued job class.
     *
     * Resolves the name of "wrapped" jobs such as class-based handlers.
     *
     * @return string
     */
    public function resolveName(): string
    {
        return CompatJobName::resolve($this->getName(), $this->payload());
    }

    /**
     * Get the name of the queue the job belongs to.
     *
     * @return string
     */
    public function getQueue(): string
    {
        return $this->queue;
    }

    /**
     * Get the service container instance.
     *
     * @return Container
     */
    public function getContainer(): Container
    {
        return $this->container;
    }

    /**
     * Get the number of times the job has been attempted.
     *
     * @return int
     */
    public function attempts(): int
    {
        return data_get($this->payload(), 'attempts', 1);
    }

    /**
     * Get the raw body string for the job.
     *
     * @return string
     */
    public function getRawBody(): string
    {
        return json_encode($this->payload);
    }

    /**
     * Get the job identifier.
     *
     * @return string
     */
    public function getJobId()
    {
        return data_get($this->payload(), 'id') ?? '';
    }
}
